package
{
	import away3d.cameras.Camera3D;
	import away3d.containers.Scene3D;
	import away3d.containers.View3D;
	import away3d.debug.AwayStats;
	import away3d.entities.Mesh;
	import away3d.lights.DirectionalLight;
	import away3d.lights.PointLight;
	import away3d.lights.shadowmaps.DirectionalShadowMapper;
	import away3d.materials.TextureMultiPassMaterial;
	import away3d.materials.lightpickers.StaticLightPicker;
	import away3d.materials.methods.FresnelSpecularMethod;
	import away3d.materials.methods.GradientDiffuseMethod;
	import away3d.materials.methods.SoftShadowMapMethod;
	import away3d.textures.SpecularBitmapTexture;
	import away3d.utils.Cast;

	import com.derschmale.away3d.MouseHoverController;
	import com.derschmale.away3d.headShading.ConfigPanel;
	import com.derschmale.away3d.headShading.ParseMeshTask;
	import com.derschmale.away3d.materials.methods.WorldNormalMethod;
	import com.derschmale.patterns.tasks.Task;

	import flash.display.Sprite;
	import flash.events.Event;
	import flash.events.MouseEvent;

	[SWF(frameRate="60", backgroundColor="#000000")]
	[Frame(factoryClass="com.derschmale.away3d.headShading.Preloader")]
	public class Main extends Sprite
	{
		[Embed(source="/../embeds/head_diffuse.jpg")]
		private var Albedo : Class;

		[Embed(source="/../embeds/head_normals_world.jpg")]
		private var Normals : Class;

		[Embed(source="/../embeds/head_specular.jpg")]
		private var Specular : Class;

		[Embed(source="/../embeds/diffuseGradient.jpg")]
		private var DiffuseGradient : Class;

		private var _bootstrap : ParseMeshTask;

		private var view : View3D;
		private var camera : Camera3D;
		private var scene : Scene3D;

		private var directionalLight : DirectionalLight;
		private var blueLight : PointLight;
		private var redLight : PointLight;

		private var cameraController : MouseHoverController;

		private var headModel : Mesh;
		private var headMaterial : TextureMultiPassMaterial;
		private var softShadows : SoftShadowMapMethod;

		private var configPanel : ConfigPanel;

		public function Main()
		{
			if (stage)
				onAddedToStage(null);
			else
				addEventListener(Event.ADDED_TO_STAGE, onAddedToStage);
		}

		/**
		 * Provides asynchronous task to perform during startup (so Preloader can show it)
		 */
		public function get bootStrap() : Task
		{
			return _bootstrap ||= new ParseMeshTask();
		}

		private function onAddedToStage(event : Event) : void
		{
			removeEventListener(Event.ADDED_TO_STAGE, onAddedToStage);
			initView();
			initCamera();
			initLights();
			initMaterial();
			initMesh();
			initUI();
			onResize();

			stage.addEventListener(Event.RESIZE, onResize);
			addEventListener(Event.ENTER_FRAME, onEnterFrame);
		}

		/**
		 * Updates camera and renders the scene every frame
		 */
		private function onEnterFrame(event : Event) : void
		{
			cameraController.update();
			view.render();
		}

		/**
		 * Preps the head material
		 */
		private function initMaterial() : void
		{
			// be sure to use multipass shading for better shadows/lights
			headMaterial = new TextureMultiPassMaterial(Cast.bitmapTexture(Albedo));
			headMaterial.normalMap = Cast.bitmapTexture(Normals);
			headMaterial.normalMethod = new WorldNormalMethod();
			headMaterial.lightPicker = new StaticLightPicker([directionalLight, blueLight, redLight]);

			// create soft shadows with a lot of samples for best results. With the current method setup, any more samples
			// would fail to compile
			softShadows = new SoftShadowMapMethod(directionalLight, 29);
			softShadows.range = 3;	// the sample radius, defines the softness of the shadows
			softShadows.epsilon = .005;
			headMaterial.shadowMethod = softShadows;

			// create specular reflections that are stronger from the sides
			var fresnel : FresnelSpecularMethod = new FresnelSpecularMethod(true);
			fresnel.fresnelPower = 3;
			headMaterial.specularMethod = fresnel;
			headMaterial.specularMap = new SpecularBitmapTexture(Cast.bitmapData(Specular));
			headMaterial.specular = 3;
			headMaterial.gloss = 10;

			// very low-cost and crude subsurface scattering for diffuse shading
			headMaterial.diffuseMethod = new GradientDiffuseMethod(Cast.bitmapTexture(DiffuseGradient));
		}

		private function initView() : void
		{
			view = new View3D();
			camera = view.camera;
			scene = view.scene;
			addChild(view);
		}

		private function initCamera() : void
		{
			// don't need much more
			camera.lens.near = 20;
			camera.lens.far = 400;
			camera.y = 0;
			camera.z = -200;
			cameraController = new MouseHoverController(stage, camera, null);
		}

		private function initLights() : void
		{
			var initialAzimuth : Number = .6;
			var initialArc : Number = 2;
			var x : Number = Math.sin(initialAzimuth)*Math.cos(initialArc);
			var y : Number = -Math.cos(initialAzimuth);
			var z : Number = Math.sin(initialAzimuth)*Math.sin(initialArc);

			// main light casting the shadows
			directionalLight = new DirectionalLight(x, y, z);
			directionalLight.color = 0xffeedd;
			directionalLight.ambient = 1;
			directionalLight.specular = .3;
			directionalLight.ambientColor = 0x101025;
			directionalLight.castsShadows = true;
			DirectionalShadowMapper(directionalLight.shadowMapper).lightOffset = 1000;
			scene.addChild(directionalLight);

			// blue point light coming from the right
			blueLight = new PointLight();
			blueLight.color = 0x4080ff;
			blueLight.x = 3000;
			blueLight.z = 700;
			blueLight.y = 20;
			scene.addChild(blueLight);

			// red light coming from the left
			redLight = new PointLight();
			redLight.color = 0x802010;
			redLight.x = -2000;
			redLight.z = 800;
			redLight.y = -400;
			scene.addChild(redLight);
		}

		/**
		 * Simply assigns the material and adds the mesh to the stage
		 */
		private function initMesh() : void
		{
			headModel = _bootstrap.headModel;
			headModel.material = headMaterial;
			headModel.geometry.scale(25);
			headModel.y = -20;
			scene.addChild(headModel);
		}

		private function initUI() : void
		{
			configPanel = new ConfigPanel(directionalLight, softShadows);
			configPanel.addEventListener(MouseEvent.MOUSE_DOWN, blockMouseDown);
			addChild(configPanel);
		}

		private function blockMouseDown(event : MouseEvent) : void
		{
			// prevent interaction with UI to trigger interaction updates
			event.stopImmediatePropagation();
		}

		private function onResize(event : Event = null) : void
		{
			view.width = stage.stageWidth;
			view.height = stage.stageHeight;
			configPanel.x = stage.stageWidth - configPanel.width;
		}
	}
}
